<?php

/**
 * Pimcore
 *
 * This source file is available under two different licenses:
 * - GNU General Public License version 3 (GPLv3)
 * - Pimcore Commercial License (PCL)
 * Full copyright and license information is available in
 * LICENSE.md which is distributed with this source code.
 *
 *  @copyright  Copyright (c) Pimcore GmbH (http://www.pimcore.org)
 *  @license    http://www.pimcore.org/license     GPLv3 and PCL
 */

namespace Pimcore\Model\DataObject\QuantityValue;

use Pimcore\Cache;
use Pimcore\Event\DataObjectQuantityValueEvents;
use Pimcore\Event\Model\DataObject\QuantityValueUnitEvent;
use Pimcore\Event\Traits\RecursionBlockingEventDispatchHelperTrait;
use Pimcore\Model;

/**
 * @method \Pimcore\Model\DataObject\QuantityValue\Unit\Dao getDao()
 */
class Unit extends Model\AbstractModel
{
    use RecursionBlockingEventDispatchHelperTrait;

    const CACHE_KEY = 'quantityvalue_units_table';

    protected ?string $id = null;

    protected string $abbreviation;

    protected string $group;

    protected string $longname;

    protected ?string $baseunit = null;

    protected string $reference;

    protected ?float $factor = null;

    protected ?float $conversionOffset = null;

    protected string $converter;

    public static function getByAbbreviation(string $abbreviation): ?Unit
    {
        try {
            $unit = new self();
            $unit->getDao()->getByAbbreviation($abbreviation);

            return $unit;
        } catch (Model\Exception\NotFoundException $e) {
            return null;
        }
    }

    public static function getByReference(string $reference): ?Unit
    {
        try {
            $unit = new self();
            $unit->getDao()->getByReference($reference);

            return $unit;
        } catch (Model\Exception\NotFoundException $e) {
            return null;
        }
    }

    public static function getById(string $id): ?Unit
    {
        try {
            $table = null;
            if (Cache\RuntimeCache::isRegistered(self::CACHE_KEY)) {
                $table = Cache\RuntimeCache::get(self::CACHE_KEY);
            }

            if (!is_array($table)) {
                $table = Cache::load(self::CACHE_KEY);
                if (is_array($table)) {
                    Cache\RuntimeCache::set(self::CACHE_KEY, $table);
                }
            }

            if (!is_array($table)) {
                $table = [];
                $list = new Model\DataObject\QuantityValue\Unit\Listing();
                $list = $list->load();
                foreach ($list as $item) {
                    $table[$item->getId()] = $item;
                }

                Cache::save($table, self::CACHE_KEY, [], null, 995, true);
                Cache\RuntimeCache::set(self::CACHE_KEY, $table);
            }
        } catch (\Exception $e) {
            return null;
        }

        if (isset($table[$id])) {
            return $table[$id];
        }

        return null;
    }

    public static function create(array $values = []): Unit
    {
        $unit = new self();
        $unit->setValues($values);

        return $unit;
    }

    public function save()
    {
        $isUpdate = false;
        if ($this->getId()) {
            $isUpdate = true;
            $this->dispatchEvent(new QuantityValueUnitEvent($this), DataObjectQuantityValueEvents::UNIT_PRE_UPDATE);
        } else {
            $this->dispatchEvent(new QuantityValueUnitEvent($this), DataObjectQuantityValueEvents::UNIT_PRE_ADD);
        }

        $this->getDao()->save();
        Cache\RuntimeCache::set(self::CACHE_KEY, null);
        Cache::remove(self::CACHE_KEY);

        if ($isUpdate) {
            $this->dispatchEvent(new QuantityValueUnitEvent($this), DataObjectQuantityValueEvents::UNIT_POST_UPDATE);
        } else {
            $this->dispatchEvent(new QuantityValueUnitEvent($this), DataObjectQuantityValueEvents::UNIT_POST_ADD);
        }
    }

    public function delete()
    {
        $this->dispatchEvent(new QuantityValueUnitEvent($this), DataObjectQuantityValueEvents::UNIT_PRE_DELETE);
        $this->getDao()->delete();
        Cache\RuntimeCache::set(self::CACHE_KEY, null);
        Cache::remove(self::CACHE_KEY);
        $this->dispatchEvent(new QuantityValueUnitEvent($this), DataObjectQuantityValueEvents::UNIT_POST_DELETE);
    }

    /**
     * @return string
     */
    public function __toString()
    {
        return ucfirst($this->getAbbreviation() . ' (' . $this->getId() . ')');
    }

    public function setAbbreviation(string $abbreviation): static
    {
        $this->abbreviation = $abbreviation;

        return $this;
    }

    public function getAbbreviation(): string
    {
        return $this->abbreviation;
    }

    public function setBaseunit(Unit|string|null $baseunit): static
    {
        if ($baseunit instanceof self) {
            $baseunit = $baseunit->getId();
        }
        $this->baseunit = $baseunit;

        return $this;
    }

    public function getBaseunit(): ?Unit
    {
        if ($this->baseunit) {
            return self::getById($this->baseunit);
        }

        return null;
    }

    public function setFactor(float $factor): static
    {
        $this->factor = $factor;

        return $this;
    }

    public function getFactor(): ?float
    {
        return $this->factor;
    }

    public function setGroup(string $group): static
    {
        $this->group = $group;

        return $this;
    }

    public function getGroup(): string
    {
        return $this->group;
    }

    public function setId(string $id): static
    {
        $this->id = (string) $id;

        return $this;
    }

    public function getId(): string
    {
        return (string) $this->id;
    }

    public function setLongname(string $longname): static
    {
        $this->longname = $longname;

        return $this;
    }

    public function getLongname(): string
    {
        return $this->longname;
    }

    public function getReference(): string
    {
        return $this->reference;
    }

    public function setReference(string $reference): static
    {
        $this->reference = $reference;

        return $this;
    }

    public function getConversionOffset(): ?float
    {
        return $this->conversionOffset;
    }

    public function setConversionOffset(float $conversionOffset): static
    {
        $this->conversionOffset = $conversionOffset;

        return $this;
    }

    public function getConverter(): string
    {
        return $this->converter;
    }

    public function setConverter(string $converter): static
    {
        $this->converter = (string)$converter;

        return $this;
    }
}
